#region usings
using System;
using System.ComponentModel.Composition;
using System.Collections.Generic;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.Utils.VMath;
using VVVV.Utils.Streams;
using VVVV.Utils;
#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "Map", Category = "Value", Version = "Pointer", Help = "Maps the value in the given range to a proportional value in the given output range", Tags = "woei")]
	#endregion PluginInfo
	public class MapPointerNode : IPluginEvaluate 
	{
		#region fields & pins
		[Input("Input")]
		public IValueFastIn FInput;
		
		[Input("Bin Size", AsInt = true, DefaultValue=1, Visibility = PinVisibility.Hidden)]
		public IValueFastIn FBin;
		
		[Input("Source Minimum", DefaultValue = 0.0)]
		public IValueFastIn FSrcMin;
		
		[Input("Source Maximum", DefaultValue = 1.0)]
		public IValueFastIn FSrcMax;
		
		[Input("Destination Minimum", DefaultValue = 0.0)]
		public IValueFastIn FDstMin;
		
		[Input("Destination Maximum", DefaultValue = 1.0)]
		public IValueFastIn FDstMax;
		
		[Input("Mapping", EnumName = "MapRangeMode")]
		public ISpread<EnumEntry> FMapping;

		[Output("Output")]
		public IValueOut FOutput;
		#endregion fields & pins
		
		public unsafe void Evaluate(int spreadMax)
		{
			int cIn, cBin, cSrcMin, cSrcMax, cDstMin, cDstMax;
			double* inP, binP, srcMinP, srcMaxP, dstMinP, dstMaxP, outP;
			
			FInput.GetValuePointer(out cIn, out inP);
			FBin.GetValuePointer(out cBin, out binP);
			FSrcMin.GetValuePointer(out cSrcMin, out srcMinP);
			FSrcMax.GetValuePointer(out cSrcMax, out srcMaxP);
			FDstMin.GetValuePointer(out cDstMin, out dstMinP);
			FDstMax.GetValuePointer(out cDstMax, out dstMaxP);
			
			spreadMax = cBin.CombineStreams(cSrcMin).CombineStreams(cSrcMax)
							.CombineStreams(cDstMin).CombineStreams(cDstMax)
							.CombineStreams(FMapping.SliceCount);
			
			int binC = 0;
			int sliceMax = 0;
			List<int> bins = new List<int>();
			if (spreadMax > 0)
			{
				while (binC < spreadMax || sliceMax <cIn)
				{
					int bin = (int)binP[binC%cBin];
					if (bin<0)
						bin = (int)Math.Round(cIn/(double)Math.Abs(bin));
					sliceMax += bin;
					binC++;
					bins.Add(bin);
				}
			}
			
			FOutput.SliceCount = sliceMax;
			FOutput.GetValuePointer(out outP);
			int i=0;
			for (int b=0; b<bins.Count; b++)
			{
				double sMin = srcMinP[b%cSrcMin];
				double sMax = srcMaxP[b%cSrcMax];
				double dMin = dstMinP[b%cDstMin];
				double dMax = dstMaxP[b%cDstMax];

				for (int s=0; s<bins[b]; s++)
				{
					double val = FromRange(inP[i%cIn], sMin, sMax);
					switch (FMapping[b].Index)
					{
						case 3:
						{
							val = Math.Max(0, Math.Min(1, val));
							break;
						}
						case 2:
						{
							int frac = (int)Math.Floor(val) ;
							val-=frac;
							if ((frac & 1) == 1) val=1-val;
							break;
						}
						case 1:
						{
							val -= Math.Floor(val);
							break;
						}
					}
					outP[i] = ToRange(val, dMin, dMax);
					i++;
				}
			}
		}
		
		private double FromRange(double val, double sMin, double sMax)
		{
			double sLen = sMax-sMin;
			if (sLen == 0)
				return 0;
			else
				return  (val-sMin)/sLen; 
		}
		
		private double ToRange(double val, double dMin, double dMax)
		{
			double dLen = dMax-dMin;
			val*= dLen;
			return val + dMin;
		}
	}
	
	#region PluginInfo
	[PluginInfo(Name = "MapRange", Category = "Value", Version = "Pointer", Help = "Maps the value in the given range to a proportional value in the given output range", Tags = "woei")]
	#endregion PluginInfo
	public class MapRangePointerNode : IPluginEvaluate 
	{
		#region fields & pins
		[Input("Input")]
		public IValueFastIn FInput;
		
		[Input("Bin Size", AsInt = true, DefaultValue=1, Visibility = PinVisibility.Hidden)]
		public IValueFastIn FBin;
		
		[Input("Source Center", DefaultValue = 0.5)]
		public IValueFastIn FSrcMin;
		
		[Input("Source Width", DefaultValue = 1.0)]
		public IValueFastIn FSrcMax;
		
		[Input("Destination Center", DefaultValue = 0.5)]
		public IValueFastIn FDstMin;
		
		[Input("Destination Width", DefaultValue = 1.0)]
		public IValueFastIn FDstMax;
		
		[Input("Mapping", EnumName = "MapRangeMode")]
		public ISpread<EnumEntry> FMapping;

		[Output("Output")]
		public IValueOut FOutput;
		#endregion fields & pins
		
		public unsafe void Evaluate(int spreadMax)
		{
			int cIn, cBin, cSrcMin, cSrcMax, cDstMin, cDstMax;
			double* inP, binP, srcMinP, srcMaxP, dstMinP, dstMaxP, outP;
			
			FInput.GetValuePointer(out cIn, out inP);
			FBin.GetValuePointer(out cBin, out binP);
			FSrcMin.GetValuePointer(out cSrcMin, out srcMinP);
			FSrcMax.GetValuePointer(out cSrcMax, out srcMaxP);
			FDstMin.GetValuePointer(out cDstMin, out dstMinP);
			FDstMax.GetValuePointer(out cDstMax, out dstMaxP);
			
			spreadMax = cBin.CombineStreams(cSrcMin).CombineStreams(cSrcMax)
							.CombineStreams(cDstMin).CombineStreams(cDstMax)
							.CombineStreams(FMapping.SliceCount);
			
			int binC = 0;
			int sliceMax = 0;
			List<int> bins = new List<int>();
			if (spreadMax > 0)
			{
				while (binC < spreadMax || sliceMax <cIn)
				{
					int bin = (int)binP[binC%cBin];
					if (bin<0)
						bin = (int)Math.Round(cIn/(double)Math.Abs(bin));
					sliceMax += bin;
					binC++;
					bins.Add(bin);
				}
			}
			
			FOutput.SliceCount = sliceMax;
			FOutput.GetValuePointer(out outP);
			int i=0;
			for (int b=0; b<bins.Count; b++)
			{
				double sMin = srcMinP[b%cSrcMin];
				double sMax = srcMaxP[b%cSrcMax];
				double dMin = dstMinP[b%cDstMin];
				double dMax = dstMaxP[b%cDstMax];

				for (int s=0; s<bins[b]; s++)
				{
					double val = FromRange(inP[i%cIn], sMin, sMax);
					switch (FMapping[b].Index)
					{
						case 3:
						{
							val = Math.Max(0, Math.Min(1, val));
							break;
						}
						case 2:
						{
							int frac = (int)Math.Floor(val) ;
							val-=frac;
							if ((frac & 1) == 1) val=1-val;
							break;
						}
						case 1:
						{
							val -= Math.Floor(val);
							break;
						}
					}
					outP[i] = ToRange(val, dMin, dMax);
					i++;
				}
			}
		}
		
		private double FromRange(double val, double sMin, double sMax)
		{
			sMin-=sMax/2;
			if (sMax == 0)
				return 0;
			else
				return  (val-sMin)/sMax; 
		}
		
		private double ToRange(double val, double dMin, double dMax)
		{
			val*= dMax;
			dMin-= dMax/2;
			return val + dMin;
		}
	}
}